use crate::{
    entities::view::parser::{ViewId, *},
    errors::{ErrorBuilder, WorkspaceError, WsErrCode},
};
use flowy_derive::ProtoBuf;
use std::convert::TryInto;

#[derive(Default, ProtoBuf)]
pub struct UpdateViewRequest {
    #[pb(index = 1)]
    pub view_id: String,

    #[pb(index = 2, one_of)]
    pub name: Option<String>,

    #[pb(index = 3, one_of)]
    pub desc: Option<String>,

    #[pb(index = 4, one_of)]
    pub thumbnail: Option<String>,

    #[pb(index = 5, one_of)]
    pub is_trash: Option<bool>,
}

pub struct UpdateViewParams {
    pub view_id: String,
    pub name: Option<String>,
    pub desc: Option<String>,
    pub thumbnail: Option<String>,
    pub is_trash: Option<bool>,
}

impl TryInto<UpdateViewParams> for UpdateViewRequest {
    type Error = WorkspaceError;

    fn try_into(self) -> Result<UpdateViewParams, Self::Error> {
        let view_id = ViewId::parse(self.view_id)
            .map_err(|e| ErrorBuilder::new(WsErrCode::ViewIdInvalid).msg(e).build())?
            .0;

        let name = match self.name {
            None => None,
            Some(name) => Some(
                ViewName::parse(name)
                    .map_err(|e| ErrorBuilder::new(WsErrCode::ViewNameInvalid).msg(e).build())?
                    .0,
            ),
        };

        let desc = match self.desc {
            None => None,
            Some(desc) => Some(
                ViewDesc::parse(desc)
                    .map_err(|e| ErrorBuilder::new(WsErrCode::ViewDescInvalid).msg(e).build())?
                    .0,
            ),
        };

        let thumbnail = match self.thumbnail {
            None => None,
            Some(thumbnail) => Some(
                ViewThumbnail::parse(thumbnail)
                    .map_err(|e| {
                        ErrorBuilder::new(WsErrCode::ViewThumbnailInvalid)
                            .msg(e)
                            .build()
                    })?
                    .0,
            ),
        };

        Ok(UpdateViewParams {
            view_id,
            name,
            desc,
            thumbnail,
            is_trash: self.is_trash,
        })
    }
}
